"""Muffin Handlers."""

import inspect
import typing as t

from asgi_tools import Request
from asgi_tools.app import HTTP_METHODS, HTTPView
from asgi_tools.utils import is_awaitable
from http_router import Router
from http_router.typing import TYPE_METHODS


class HandlerMeta(type):

    """Prepare handlers."""

    def __new__(mcs, name, bases, params):
        """Prepare a Handler Class."""
        cls: Handler = super().__new__(mcs, name, bases, params)  # type: ignore

        # Ensure that the class methods are exist and iterable
        if not cls.methods:
            cls.methods = [
                method for method in HTTP_METHODS if method.lower() in cls.__dict__
            ]

        elif isinstance(cls.methods, str):
            cls.methods = [cls.methods]

        cls.methods = set(method.upper() for method in cls.methods)
        for m in cls.methods:
            method = getattr(cls, m.lower(), None)
            if method and not is_awaitable(method):
                raise TypeError(
                    f"The method '{method.__qualname__}' has to be awaitable"
                )

        return cls


def route_method(*paths: str, **params) -> t.Callable:
    """Mark a method as a route."""

    def wrapper(method):
        """Wrap a method."""
        method.__route__ = paths, params
        return method

    return wrapper


class Handler(HTTPView, metaclass=HandlerMeta):

    """Class-based view pattern for handling HTTP method dispatching.

    .. code-block:: python

        @app.route('/hello', '/hello/{name}')
        class HelloHandler(Handler):

            async def get(self, request):
                name = request.patch_params.get('name') or 'all'
                return "GET: Hello f{name}"

            async def post(self, request):
                name = request.patch_params.get('name') or 'all'
                return "POST: Hello f{name}"

            @Handler.route('/hello/custom')
            async def custom(self, request):
                return 'Custom HELLO'

        # ...
        async def test_my_endpoint(client):
            response = await client.get('/hello')
            assert await response.text() == 'GET: Hello all'

            response = await client.get('/hello/john')
            assert await response.text() == 'GET: Hello john'

            response = await client.post('/hello')
            assert await response.text() == 'POST: Hello all'

            response = await client.get('/hello/custom')
            assert await response.text() == 'Custom HELLO'

            response = await client.delete('/hello')
            assert response.status_code == 405

    """

    methods: t.Optional[t.Collection[str]] = None

    @classmethod
    def __route__(
        cls, router: Router, *paths: str, methods: TYPE_METHODS = None, **params
    ):
        """Check for registered methods."""
        router.bind(cls, *paths, methods=methods or cls.methods, **params)
        for _, method in inspect.getmembers(cls, lambda m: hasattr(m, "__route__")):
            cpaths, cparams = method.__route__
            router.bind(cls, *cpaths, __meth__=method.__name__, **cparams)

        return cls

    def __call__(self, request: Request, **opts) -> t.Awaitable:
        """Dispatch the given request by HTTP method."""
        method = getattr(self, opts.get("__meth__") or request.method.lower())
        return method(request)

    route = route_method



"""Muffin is a web framework to build ASGI applications."""

# Package information
# ===================

__version__ = "0.87.1"
__project__ = "muffin"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "MIT"

from asgi_tools import ASGINotFound  # noqa
from asgi_tools import (ASGIConnectionClosed, ASGIError, ASGIMethodNotAllowed, Request, Response,
                        ResponseError, ResponseFile, ResponseHTML, ResponseJSON, ResponseRedirect,
                        ResponseSSE, ResponseStream, ResponseText, ResponseWebSocket)
from asgi_tools.tests import ASGITestClient as TestClient  # noqa


class MuffinException(ASGIError):

    """Base class for Muffin Errors."""

    pass


CONFIG_ENV_VARIABLE = "MUFFIN_CONFIG"


from .app import Application  # noqa
from .handler import Handler  # noqa

__all__ = (
    "ASGIConnectionClosed",
    "ASGIError",
    "ASGIMethodNotAllowed",
    "ASGINotFound",
    "Application",
    "Handler",
    "Request",
    "Response",
    "ResponseError",
    "ResponseFile",
    "ResponseHTML",
    "ResponseJSON",
    "ResponseRedirect",
    "ResponseSSE",
    "ResponseStream",
    "ResponseText",
    "ResponseWebSocket",
    "TestClient",
)



"""A helper to write Muffin Plugins."""

import typing as t
from abc import ABC

from asgi_tools.utils import to_awaitable
from modconfig import Config

from muffin import MuffinException
from muffin.app import Application


class PluginException(MuffinException):

    """Implement any exception in plugins."""


class BasePlugin(ABC):

    """Base class for Muffin plugins."""

    name: str

    app: t.Optional[Application] = None

    # Plugin options with default values
    defaults: t.Dict[str, t.Any] = {}

    # Optional middleware method
    middleware: t.Optional[t.Callable] = None

    # Optional startup method
    startup: t.Optional[t.Callable] = None

    # Optional shutdown method
    shutdown: t.Optional[t.Callable] = None

    # Optional conftest method
    conftest: t.Optional[t.Callable] = None

    def __init__(self, app: Application = None, **options):
        """Save application and create he plugin's configuration."""
        if getattr(self, "name", None) is None:
            raise TypeError("Plugin.name is required")

        self.cfg = Config(config_config={"update_from_env": False}, **self.defaults)

        if app is not None:
            self.setup(app, **options)
        else:
            self.cfg.update_from_dict(options, exist_only=True)

    def __repr__(self) -> str:
        """Human readable representation."""
        return f"<muffin.Plugin: { self.name }>"

    @property
    def installed(self):
        """Check the plugin is installed to an app."""
        return bool(self.app)

    def setup(self, app: Application, *, name: str = None, **options):
        """Bind app and update the plugin's configuration."""
        # allow to redefine the name for multi plugins with same type
        self.name = name or self.name  # type: ignore
        self.app = app
        self.app.plugins[self.name] = self

        # Update configuration
        self.cfg.update_from_dict(
            dict(app.cfg), prefix=f"{self.name}_", exist_only=True
        )
        self.cfg.update_from_dict(options)

        # Register a middleware
        if self.middleware:
            self.app.middleware(to_awaitable(self.middleware))

        # Bind startup
        if self.startup:
            self.app.on_startup(self.startup)

        # Bind shutdown
        if self.shutdown:
            self.app.on_shutdown(self.shutdown)
